import 'dart:ffi';
import 'package:meta/meta.dart';
import 'array.dart';

/// Instead of doing many little allocations, [Arena]
/// allows you to make one large allocation which can be sliced into multiple [ArrayView]s.
/// This can significantly improve performance and it makes
/// memory management simpler.
/// ````
/// final buffer = Array<Uint8>(10);
/// final arena = Arena(buffer);
/// final array = arena.newArray(5);
/// final array2 = arena.newArray(5);
/// void some_native_c_api(array.ptr, array.length, array2.ptr, array2.length);
/// arena.buffer.delete();
/// ````
class Arena<T extends NativeType> {
  int _index = 0;

  final Array<T> buffer;

  /// Left free space of Arena.
  @nonVirtual
  int get freeSpaceLeft => buffer.length - _index;

  /// Allocates a buffer of [length]
  Arena(this.buffer);

  /// Returns a view into [buffer].
  /// When [length] exceeds [freeSpaceLeft] it throws a [RangeError].
  ArrayView<T> newArray(int length) {
    assert(length > 0);
    if (freeSpaceLeft < length) {
      throw RangeError('Not enough free space');
    }
    final ptrAtIndex = buffer.ptr.elementAt(_index);
    final arr = ArrayView<T>.fromPtr(ptrAtIndex, length);
    _index += length;
    return arr;
  }
}
