import 'dart:ffi';
import 'dart:typed_data';
import 'package:ffi/ffi.dart' as ffi;
import 'package:meta/meta.dart';

/// Stores the [ptr] to an array and its [length]
mixin Array<T extends NativeType> {
  Pointer<T> get ptr;
  int get length;

  /// Frees [ptr] to the array.
  void delete();

  @override
  String toString() {
    return '${runtimeType}($ptr, $length)';
  }
}

/// Basic implemenatation of [Array] using [ffi](https://pub.dev/packages/ffi)
/// for allocation and deallocation
class BasicArray<T extends NativeType> with Array<T> {
  @override
  final Pointer<T> ptr;
  @override
  final int length;
  BasicArray(this.length) : ptr = ffi.allocate(count: length);
  BasicArray.fromPtr(this.ptr, this.length);

  @override
  void delete() {
    ffi.free(ptr);
  }
}

/// Non-owning [Array], meaning [delete] doesn't do anything.
class ArrayView<T extends NativeType> with Array<T> {
  @override
  final Pointer<T> ptr;
  @override
  final int length;
  ArrayView.fromPtr(this.ptr, this.length);
  ArrayView.fromArray(Array<T> arr)
      : ptr = arr.ptr,
        length = arr.length;
  @override
  void delete() {}
}

/// Takes an [Array] and exposes a [list] to its contents.
/// The exact type of list can be narrowed with extension methods like [Uint8ArrayList].
/// [list] becomes invalid after [array] got deleted.
class ArrayList<T extends NativeType, Arr extends Array<T>> {
  @visibleForOverriding
  final List<int> list;
  final Arr array;

  /// [list] is a view into the contents of [arr.ptr]; see extension methods of [Pointer] at
  /// [dart:ffi](https://api.dart.dev/stable/2.10.5/dart-ffi/dart-ffi-library.html).
  /// To create an [ArrayList], prefer extension methods like [AsUint8Arraylist].
  ArrayList(this.array, this.list);
}

extension AsUint8ArrayList on Array<Uint8> {
  @factory
  ArrayList<Uint8, Array<Uint8>> asArrayList() {
    return ArrayList(this, ptr.asTypedList(length));
  }
}

extension Uint8ArrayList on ArrayList<Uint8, Array<Uint8>> {
  Uint8List get typedList => list;
}
