import 'package:meta/meta.dart';
import 'array.dart';
import 'dart:ffi';

/// Reference counted Array. Not thread safe!
class SharedArray<T extends NativeType> with Array<T> {
  int _counter = 0;
  final Array<T> _array;
  @override
  Pointer<T> get ptr => _array.ptr;
  @override
  int get length => _array.length;
  @nonVirtual
  int get counter => _counter;
  @nonVirtual
  bool get deleted => counter == -1;

  /// Takes ownership of [arr]
  SharedArray(this._array);

  /// Returns [this] and updates [counter]. Throws [StateError]
  /// when [this] has already been deleted or when [counter] overflows.
  SharedArray<T> newShared() {
    if (deleted) {
      throw StateError(
          'Cannot create new new SharedArray from deleted SharedArray');
    }
    _counter++;
    if (counter.isNegative) {
      throw StateError(
          'Cannot create any more SharedArray. Counter reached maximum integer value');
    }
    return this;
  }

  /// Deletes the shared [Array], even with `[counter] > 0`. Calling this method multiple times does nothing.
  @mustCallSuper
  void forceDelete() {
    if (deleted) return;
    _counter = -1;
    _array.delete();
  }

  /// Deletes [_array] when [counter] has reached `0`. Calling this method multiple times
  /// does nothing.
  @override
  @mustCallSuper
  void delete() {
    if (deleted) return;
    _counter--;
    if (counter == 0) {
      _array.delete();
      _counter = -1;
    }
  }
}
